from detrain.tp.sequence_train import sequence_train_loop
def train_loop(dataloader, model_2d, loss_fn, optimizer, batch_size, device):
    sequence_train_loop(dataloader, model_2d, loss_fn, optimizer, batch_size, device)


import torch
import torch.nn as nn
import time
import os
from detrain.ppl.args_util import get_args
from detrain.tp.train_eval import train_eval
from detrain.tp.model_utils import get_tp_model, save_model
from detrain.ppl.dataset_util import get_torchvision_dataset
import torch.optim as optim
from base_model import NeuralNetwork
from torch.distributed.tensor.parallel import (
    ColwiseParallel,
    RowwiseParallel,
)
from torch.distributed._tensor import Shard

if __name__=="__main__":
    args = get_args()
    world_size = int(os.environ["WORLD_SIZE"])
    # Get args
    epochs = int(args.epochs)
    batch_size = int(args.batch_size)
    lr = float(args.lr)
    device = "cpu"

    # Check devices
    if (args.gpu is not None):
        device = "cuda"
    
    # Define optimizer & loss_fn
    loss_fn = nn.CrossEntropyLoss(reduction="mean")
    optimizer_class = optim.SGD
    model = NeuralNetwork().to(device)

   
    mesh_shape = (world_size, )
    tp_model = get_tp_model(model, {
        "in_proj": ColwiseParallel(
            use_local_output=False,
        ),
        "linear1": RowwiseParallel(
            use_local_output=False,
        ),
        "out_proj": ColwiseParallel(

            use_local_output=False,
        ),
    } , device, mesh_shape)

    
    # Create an optimizer for the parallelized module.
    optimizer = torch.optim.SGD(tp_model.parameters(), lr=lr)
    
    # Dataloaders

    (train_dataloader, test_dataloader) = get_torchvision_dataset("MNIST", batch_size)

    tik = time.time()
    train_eval(
        tp_model, 
        train_dataloader, 
        test_dataloader, 
        loss_fn, 
        optimizer, 
        epochs, 
        batch_size,
        device
    )
    tok = time.time()
    print(f"Execution time = {tok - tik}")
    save_model(model, args.model_name)


import torch.nn as nn
import torch
import time
import os
from detrain.ppl.args_util import get_args
from detrain.ppl.worker import run_worker
from detrain.ppl.dataset_util import get_torchvision_dataset
from shards_model import NNShard1, NNShard2
from torch.distributed.optim.optimizer import DistributedOptimizer
import torch.optim as optim
# Define model here
# torchrun \
# --nproc_per_node=1 --nnodes=3 --node_rank=0 \
# --master_addr=localhost --master_port=9999 \
# main.py --use_syn
if __name__=="__main__":
    args = get_args()
    # Get args
    world_size = int(os.environ["WORLD_SIZE"])
    rank = int(os.environ["RANK"])
    epochs = int(args.epochs)
    batch_size = int(args.batch_size)
    lr = float(args.lr)
    device = "cpu"

    # Check devices
    if (args.gpu is not None):
        device = "cuda"

    for i in range(torch.cuda.device_count()):
        print(torch.cuda.get_device_properties(i).name)
    

    # Define optimizer & loss_fn
    loss_fn = nn.CrossEntropyLoss()
    optimizer_class = optim.SGD
    
    # Dataloaders

    (train_dataloader, test_dataloader) = get_torchvision_dataset("MNIST", batch_size)

    
    print(f"World_size: {world_size}, Rank: {rank}")
    num_split = 4
    tik = time.time()
    run_worker(
        rank, 
        world_size, 
        (
            args.split_size, 
            ["worker1", "worker2"],
            [device, device], 
            [NNShard1, NNShard2]
        ), 
        train_dataloader, 
        test_dataloader, 
        loss_fn, 
        optimizer_class, 
        epochs, 
        batch_size,
        lr
    )
    tok = time.time()
    print(f"number of splits = {num_split}, execution time = {tok - tik}")


import torch
from detrain.ppl.train import train_loop
from detrain.ppl.evaluation import test_loop
from detrain.ppl.args_util import get_args

def run_master(model, train_dataloader, test_dataloader, loss_fn, optimizer, epochs, batch_size):
    args = get_args()
    # put the two model parts on worker1 and worker2 respectively
    # Validate master node here
    for t in range(epochs):
        print(f"Epoch {t+1}\n-------------------------------")
        train_loop(train_dataloader, model, loss_fn, optimizer, batch_size)
        test_loop(test_dataloader, model, loss_fn)
    # Save model
    states = model.state_dict()
    torch.save(states, f"{args.model_name}.pt")
    print("Done!")




import torch.nn as nn
import time
from detrain.ppl.args_util import get_args
from detrain.ppl.worker import run_worker
from detrain.ppl.dis_model import DistributedModel
from detrain.ppl.dataset_util import get_torchvision_dataset
from shards_model import ResNetShard1, ResNetShard2
from torch.distributed.optim.optimizer import DistributedOptimizer
import torch.optim as optim
# Define model here

if __name__=="__main__":
    args = get_args()

    # Get args
    world_size = int(args.world_size)
    rank = int(args.rank)
    epochs = int(args.epochs)
    batch_size = int(args.batch_size)
    device = "cpu"

    # Check devices
    if (args.gpu is not None):
        device = "cuda:0"

    model = DistributedModel(
        args.split_size, 
        ["worker1", "worker2"],
        [device, device], 
        [ResNetShard1, ResNetShard2]
    )
    
    # Define optimizer & loss_fn
    loss_fn = nn.MSELoss()
    optimizer = DistributedOptimizer(
        optim.SGD,
        model.parameter_rrefs(),
        lr=args.lr,
    )
    # Dataloaders

    (train_dataloader, test_dataloader) = get_torchvision_dataset("M", batch_size)

    
    print(f"World_size: {world_size}, Rank: {rank}")
    num_split = 4
    tik = time.time()
    run_worker(rank, world_size, model, train_dataloader, test_dataloader, loss_fn, optimizer, epochs, batch_size)
    tok = time.time()
    print(f"number of splits = {num_split}, execution time = {tok - tik}")


import torch.distributed.autograd as dist_autograd

def train_loop(dataloader, model, loss_fn, optimizer, batch_size):
    size = len(dataloader.dataset)
    # Set the model to training mode - important for batch normalization and dropout layers
    # Unnecessary in this situation but added for best practices
    model.train()
    for batch, (X, y) in enumerate(dataloader):
        # The distributed autograd context is the dedicated scope for the
        # distributed backward pass to store gradients, which can later be
        # retrieved using the context_id by the distributed optimizer.
        with dist_autograd.context() as context_id:
            pred = model(X)
            loss = loss_fn(pred, y)
            dist_autograd.backward(context_id, [loss])
            optimizer.step(context_id)
        if batch % 100 == 0:
            loss, current = loss.item(), batch * batch_size + len(X)
            print(f"loss: {loss:>7f}  [{current:>5d}/{size:>5d}]")


from detrain.tp.sequence_train import sequence_train_loop
from detrain.tp.sequence_evaluation import test_loop
import os
def sequence_train_eval(tp_model, train_dataloader, test_dataloader, loss_fn, optimizer, epochs, batch_size, device):
    rank = int(os.environ["RANK"])
    for t in range(epochs):
        print(f"\nRank{rank} -- Epoch {t+1} start\n-------------------------------")
        sequence_train_loop(train_dataloader, tp_model, loss_fn, optimizer, batch_size, device, rank)
        test_loop(test_dataloader, tp_model, loss_fn, device, rank)
    print("Done!")


import torch.nn as nn
import threading
from torch.distributed.rpc import RRef

num_classes = 1000
def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)

class ResNetBase(nn.Module):
    def __init__(self, block, inplanes, num_classes=num_classes,
                 groups=1, width_per_group=64, norm_layer=None):
        super(ResNetBase, self).__init__()

        self._lock = threading.Lock()
        self._block = block
        self._norm_layer = nn.BatchNorm2d
        self.inplanes = inplanes
        self.dilation = 1
        self.groups = groups
        self.base_width = width_per_group

    def _make_layer(self, planes, blocks, stride=1):
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if stride != 1 or self.inplanes != planes * self._block.expansion:
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes * self._block.expansion, stride),
                norm_layer(planes * self._block.expansion),
            )

        layers = []
        layers.append(self._block(self.inplanes, planes, stride, downsample, self.groups,
                                  self.base_width, previous_dilation, norm_layer))
        self.inplanes = planes * self._block.expansion
        for _ in range(1, blocks):
            layers.append(self._block(self.inplanes, planes, groups=self.groups,
                                      base_width=self.base_width, dilation=self.dilation,
                                      norm_layer=norm_layer))

        return nn.Sequential(*layers)

    def parameter_rrefs(self):
        r"""
        Create one RRef for each parameter in the given local module, and return a
        list of RRefs.
        """
        return [RRef(p) for p in self.parameters()]


from torch.utils.data import DataLoader
from torchvision import datasets
from torchvision.transforms import ToTensor
import torchvision.transforms as transforms

def get_torchvision_dataset(name, batch_size, distributed=False):
    if (name == "MNIST" and distributed == False):
        training_data = datasets.FashionMNIST(
        root="data",
        train=True,
        download=True,
        transform=ToTensor()
        )

        test_data = datasets.FashionMNIST(
            root="data",
            train=False,
            download=True,
            transform=ToTensor()
        )

        train_dataloader = DataLoader(training_data, batch_size=batch_size)
        test_dataloader = DataLoader(test_data, batch_size=batch_size)

        return (train_dataloader, test_dataloader)
    elif (name == "ImageNet" and distributed == False):
        #Prepare transformations for data augmentation
        transform = transforms.Compose([
            transforms.Resize(256),
            transforms.RandomHorizontalFlip(),
            transforms.RandomVerticalFlip(),
            transforms.RandomRotation(degrees=45),
            transforms.ColorJitter(brightness=0.5, contrast=0.5, saturation=0.5, hue=0.5),
            transforms.CenterCrop(224),
            transforms.ToTensor(),
            transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
        ])
        train_dataset = datasets.ImageNet(
            root='data', 
            split="train",
            transform=transform
        )
        test_dataset = datasets.ImageNet(
            root='data', 
            split="eval",
            transform=transform
        )
        train_dataloader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True, num_workers=2)
        test_dataloader = DataLoader(test_dataset, batch_size=batch_size, shuffle=True, num_workers=2)
        return (train_dataloader, test_dataloader)
    elif (name == "MNIST" and distributed == True):
        training_data = datasets.FashionMNIST(
        root="data",
        train=True,
        download=True,
        transform=ToTensor()
        )

        test_data = datasets.FashionMNIST(
            root="data",
            train=False,
            download=True,
            transform=ToTensor()
        )

        train_dataloader = DataLoader(training_data, batch_size=batch_size)
        test_dataloader = DataLoader(test_data, batch_size=batch_size)

        return (train_dataloader, test_dataloader)
         




import torch
import torch.nn as nn
import torch.distributed.rpc as rpc
from torch.distributed.rpc import RRef

class DistributedModel(nn.Module):
    """
    Assemble multi parts as an nn.Module and define pipelining logic
    """
    def __init__(self, split_size, workers, devices, model_shards, *args, **kwargs):
        super(DistributedModel, self).__init__()

        assert len(workers) == len(devices) and len(workers) == len(model_shards)

        self.split_size = split_size
        self.rrefs = []
        for w in range(len(workers)):
            rref = rpc.remote(
                workers[w],
                model_shards[w],
                args = (devices[w],) + args,
                **kwargs
            )
            self.rrefs.append(rref)
            
    def forward(self, xs):
        # Split the input batch xs into micro-batches, and collect async RPC
        # futures into a list
        out_futures = []
        for x in iter(xs.split(self.split_size, dim=0)):
            x_rref = RRef(x)
            y_rref = self.rrefs[0].remote().forward(x_rref)
            for i in range(len(self.rrefs)):
                if i != 0:       
                    y_rref = self.rrefs[i].rpc_async().forward(y_rref)
                    out_futures.append(y_rref)

        # collect and cat all output tensors into one tensor.
        return torch.cat(torch.futures.wait_all(out_futures))

    def parameter_rrefs(self):
        remote_params = []
        for i in range(len(self.rrefs)):
            remote_params.extend(self.rrefs[i].remote().parameter_rrefs().to_here())
        return remote_params



from torch.distributed.tensor.parallel import loss_parallel

def train_loop(dataloader, tp_model, loss_fn, optimizer, batch_size, device, rank):
    tp_model.train()
    size = len(dataloader.dataset)
    for batch, (X, y) in enumerate(dataloader):
        
        X, y = X.to(device), y.to(device)
        pred = tp_model(X)
        with loss_parallel():
            loss = loss_fn(pred, y)
            loss.backward()
            optimizer.step()
            optimizer.zero_grad()
            
        if (rank == 0):
            if batch % 100 == 0:
                loss, current = loss.item(), batch * batch_size + len(X)
                print(f"loss: {loss:>7f}  [{current:>5d}/{size:>5d}]")


import torch
from torch.distributed._tensor import DTensor
from torch.distributed.tensor.parallel import loss_parallel
import torch.distributed as dist
def test_loop(dataloader, tp_model, loss_fn, device, rank):
    # Set the model to evaluation mode - important for batch normalization and dropout layers
    # Unnecessary in this situation but added for best practices
    tp_model.eval()
    size = len(dataloader.dataset)
    tp_loss = torch.zeros(3).to(device)
    # Evaluating the model with torch.no_grad() ensures that no gradients are computed during test mode
    # also serves to reduce unnecessary gradient computations and memory usage for tensors with requires_grad=True
    with torch.no_grad():
        for X, y in dataloader:
            X, y = X.to(device), y.to(device)
            with loss_parallel():
                pred = tp_model(X)
                # Loss
                tp_loss[0] += loss_fn(pred, y).item()
                # Correct
                tp_loss[1] += (pred.argmax(1) == y).type(torch.float).sum().item()

                # Batches
                tp_loss[2] += 1
        dist.all_reduce(tp_loss, op=dist.ReduceOp.SUM)
    if (rank == 0):
        test_loss = tp_loss[0] / tp_loss[2]
        correct = tp_loss[1] / size
        print('\nTest set: Average loss: {:.4f}, Accuracy: {}/{} ({:.2f}%)\n'.format(
            test_loss, int(tp_loss[1]), int(size),
            100. * correct))


from detrain.fsdp_tp.mesh_utils import get_2d_mesh
from torch.distributed.tensor.parallel import parallelize_module
from torch.distributed.fsdp import FullyShardedDataParallel as FSDP
def get_model_2d(model, tp_plan, device_type, tp_size):
    (dp_mesh, tp_mesh) = get_2d_mesh(device_type, tp_size)
    model_tp = parallelize_module(model, tp_mesh, tp_plan)
    model_2d = FSDP(model_tp, device_mesh=dp_mesh, use_orig_params=False, device_id=device_type)
    return model_2d


import torch.distributed as dist
from torch.distributed.elastic.rendezvous import c10d_rendezvous_backend
from torch.distributed.elastic.rendezvous import dynamic_rendezvous
store = dist.TCPStore("localhost", 1522)

backend = c10d_rendezvous_backend.C10dRendezvousBackend(store, "101")



if __name__=="__main__":
    
    rdzv_handler = dynamic_rendezvous.DynamicRendezvousHandler.from_backend(
        run_id="101",
        store=store,
        backend=backend,
        min_nodes=2,
        max_nodes=4                                                                             
    )



import torch.distributed.rpc as rpc
from detrain.ppl.master_node import run_master
from detrain.ppl.dis_model import DistributedModel
from torch.distributed.optim.optimizer import DistributedOptimizer

def run_worker(rank, world_size, model_params, train_dataloader, test_dataloader, loss_fn, optimized_class, epochs, batch_size, lr):
    # Higher timeout is added to accommodate for kernel compilation time in case of ROCm.
    options = rpc.TensorPipeRpcBackendOptions(num_worker_threads=256, rpc_timeout=300)
    
    if rank == 0:
        print("--- Init master RPC")
        rpc.init_rpc(
            "master",
            rank=rank,
            world_size=world_size,
            rpc_backend_options=options
        )
        print("--- Done init master")
        model = DistributedModel(
            # split size
            model_params[0],
            # workers
            model_params[1],
            # Devices
            model_params[2],
            # Shards
            model_params[3]
        )
        optimizer = DistributedOptimizer(
            optimized_class,
            model.parameter_rrefs(),
            lr=lr,
        )
        run_master(model, train_dataloader, test_dataloader, loss_fn, optimizer, epochs, batch_size)
    else:
        print(f"--- Init worker {rank} RPC")
        rpc.init_rpc(
            f"worker{rank}",
            rank=rank,
            world_size=world_size,
            rpc_backend_options=options
        )
        print(f"--- Start to listen & receive the forwarded data from the master node")
        pass

    # block until all rpcs finish
    rpc.shutdown()



import torch.nn as nn
class NeuralNetwork(nn.Module):
    def __init__(self):
        super().__init__()
        self.flatten = nn.Flatten()
        self.in_proj = nn.Linear(28 * 28, 512)
        self.relu1 = nn.ReLU()
        self.linear1 = nn.Linear(512, 512)
        self.relu2 = nn.ReLU()
        self.out_proj = nn.Linear(512, 10)
        

    def forward(self, x):
        x = self.flatten(x)
        x = self.in_proj(x)
        x = self.relu1(x)
        x = self.linear1(x)
        x = self.relu2(x)
        logits = self.out_proj(x)
        return logits


from torch.distributed._tensor.device_mesh import init_device_mesh

from torch.distributed.tensor.parallel import (
    parallelize_module
)
import torch.distributed as dist
import torch
import os

def get_tp_model(model, parallelize_plan, device_type, mesh_shape):
    device_mesh = init_device_mesh(device_type=device_type, mesh_shape=mesh_shape)
    model = model.to(device_type)
    tp_model = parallelize_module(
        module=model,
        device_mesh=device_mesh,
        parallelize_plan=parallelize_plan,
    )
    return tp_model

def save_model(model, name):
    rank = int(os.environ["RANK"])
    dist.barrier()
    states = model.state_dict()
    if rank == 0:
        torch.save(states, f"{name}.pt")


import torch
from torch.distributed.tensor.parallel import loss_parallel
import torch.distributed as dist
def test_loop(dataloader, model_2d, loss_fn, device, rank):
    # Set the model to evaluation mode - important for batch normalization and dropout layers
    # Unnecessary in this situation but added for best practices
    model_2d.eval()
    model_2d_loss = torch.zeros(4).to(device)
    # Evaluating the model with torch.no_grad() ensures that no gradients are computed during test mode
    # also serves to reduce unnecessary gradient computations and memory usage for tensors with requires_grad=True
    with torch.no_grad():
        for X, y in dataloader:
            X, y = X.to(device), y.to(device)
            with loss_parallel():
                pred = model_2d(X)
                # Loss
                model_2d_loss[0] += loss_fn(pred, y).item()
                # Correct
                model_2d_loss[1] += (pred.argmax(1) == y).type(torch.float).sum().item()

                # Batches
                model_2d_loss[2] += 1
                # Size
                model_2d_loss[3] += len(X)
        dist.all_reduce(model_2d_loss, op=dist.ReduceOp.SUM)
    if (rank == 0):
        test_loss = model_2d_loss[0] / model_2d_loss[2]
        correct = model_2d_loss[1] / model_2d_loss[3]
        print('\nTest set: Average loss: {:.4f}, Accuracy: {}/{} ({:.2f}%)\n'.format(
            test_loss, int(model_2d_loss[1]), int(model_2d_loss[3]),
            100. * correct))


import torch
import torch.nn as nn
from torchvision.models.resnet import Bottleneck
from base_model import ResNetBase
num_classes = 1000

class ResNetShard1(ResNetBase):
    """
    The first part of ResNet.
    """
    def __init__(self, device, *args, **kwargs):
        super(ResNetShard1, self).__init__(
            Bottleneck, 64, num_classes=num_classes, *args, **kwargs)

        self.device = device
        self.seq = nn.Sequential(
            nn.Conv2d(3, self.inplanes, kernel_size=7, stride=2, padding=3, bias=False),
            self._norm_layer(self.inplanes),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=3, stride=2, padding=1),
            self._make_layer(64, 3),
            self._make_layer(128, 4, stride=2)
        ).to(self.device)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

    def forward(self, x_rref):
        x = x_rref.to_here().to(self.device)
        with self._lock:
            out =  self.seq(x)
        print(f"Forwared: worker[1]: device: {self.device}")
        return out.cpu()


class ResNetShard2(ResNetBase):
    """
    The second part of ResNet.
    """
    def __init__(self, device, *args, **kwargs):
        super(ResNetShard2, self).__init__(
            Bottleneck, 512, num_classes=num_classes, *args, **kwargs)

        self.device = device
        self.seq = nn.Sequential(
            self._make_layer(256, 6, stride=2),
            self._make_layer(512, 3, stride=2),
            nn.AdaptiveAvgPool2d((1, 1)),
        ).to(self.device)

        self.fc =  nn.Linear(512 * self._block.expansion, num_classes).to(self.device)

    def forward(self, x_rref):
        x = x_rref.to_here().to(self.device)
        with self._lock:
            out = self.fc(torch.flatten(self.seq(x), 1))
        print(f"Forwared: worker[2]: device: {self.device}")
        return out.cpu()


import torch.nn as nn  
from base_model import NeuralNetwork

class NNShard1(NeuralNetwork):
    """
    The first part of NeuralNetwork.
    """
    def __init__(self, device, *args, **kwargs):
        super(NNShard1, self).__init__(*args, **kwargs)

        self.device = device
        self.seq = nn.Sequential(
            nn.Linear(28*28, 512),
            nn.ReLU(),
            nn.Linear(512, 512),
        ).to(self.device)

    def forward(self, x_rref):
        x = x_rref.to_here().to(self.device)
        x = self.flatten(x)
        with self._lock:
            out =  self.seq(x)
        # print(f"Run forward on worker[1] - Device: {self.device}")
        return out.cpu()
    
class NNShard2(NeuralNetwork):
    """
    The first part of NeuralNetwork.
    """
    def __init__(self, device, *args, **kwargs):
        super(NNShard2, self).__init__(*args, **kwargs)

        self.device = device
        self.seq = nn.Sequential(
            nn.ReLU(),
            nn.Linear(512, 10),
        ).to(self.device)

    def forward(self, x_rref):
        x = x_rref.to_here().to(self.device)
        x = self.flatten(x)
        with self._lock:
            out =  self.seq(x)
        # print(f"Run forward on worker[2] - Device: {self.device}")
        return out.cpu()


from torch.distributed.device_mesh import init_device_mesh
import os
# For example: cuda, (dp_size,tp_size), (dp, tp)
def get_2d_mesh(device_type, tp_size):
    # Get from torchrun params
    _world_size = int(os.environ["WORLD_SIZE"])

    # Word_size = tp_size * dp_size, all numbers are integer.
    assert (
        _world_size % tp_size == 0
    ), f"World size {_world_size} needs to be divisible by TP size {tp_size}"

    dp_size = _world_size // tp_size

    # Init device mesh with two mesh names are DP, TP (DataParallelism, TensorParallelism)
    device_mesh = init_device_mesh(device_type,mesh_shape=(dp_size, tp_size), mesh_dim_names=("dp", "tp"))
    dp_mesh = device_mesh["dp"]
    tp_mesh = device_mesh["tp"]

    return (dp_mesh, tp_mesh)


import logging
import torch

logging.basicConfig(
    format="%(asctime)s %(message)s", datefmt="%m/%d/%Y %I:%M:%S %p", level=logging.INFO
)

def get_logger():
    return logging.getLogger(__name__)


def rank_log(_rank, logger, msg):
    """helper function to log only on global rank 0"""
    if _rank == 0:
        logger.info(f" {msg}")


def verify_min_gpu_count(min_gpus: int = 2) -> bool:
    """ verification that we have at least 2 gpus to run dist examples """
    has_cuda = torch.cuda.is_available()
    gpu_count = torch.cuda.device_count()
    return has_cuda and gpu_count >= min_gpus



import argparse
def get_args():
    parser = argparse.ArgumentParser(description='simple distributed training job based on pipeline parallelism')
    ## General settings
    parser.add_argument('--epochs', default=3, type=int, help='Total epochs to train the model')
    parser.add_argument('--batch_size', default=10, type=int, help='Input batch size on each device (default: 32)')
    parser.add_argument("--split_size", default=2, type=int, help="Number of parts of the input data")
    parser.add_argument('--log', default=None, type=int, 
                        help='Enable training logs, this action can affect to training process performance')
    ## Optimizer settings
    parser.add_argument('--lr', default=1e-3, type=float, help='learning rate')

    ## Cluster settings
    parser.add_argument('--gpu', default=None, type=str)
    parser.add_argument("--model_name", default="trained_model", type=str, help='Name of the trained model to be saved')
    args = parser.parse_args()
    return args



import torch.nn as nn
import threading
from torch.distributed.rpc import RRef

class NeuralNetwork(nn.Module):
    def __init__(self):
        super().__init__()
        self._lock = threading.Lock()
        self.flatten = nn.Flatten()

    def parameter_rrefs(self):
        r"""
        Create one RRef for each parameter in the given local module, and return a
        list of RRefs.
        """
        return [RRef(p) for p in self.parameters()]


import unittest
from detrain.ppl.dataset_util import get_torchvision_dataset

class TestDataset(unittest.TestCase):

    def test_imagenet_loader(self):
        (train_dataloader, test_dataloader) = get_torchvision_dataset("ImageNet", 64)
        print(train_dataloader, test_dataloader)
    def test_mnist_loader(self):
        (train_dataloader, test_dataloader) = get_torchvision_dataset("MNIST", 64)
        print(train_dataloader, test_dataloader)

if __name__=="__main__":

   unittest.main()


def sequence_train_loop(dataloader, tp_model, loss_fn, optimizer, batch_size, device, rank):
    size = len(dataloader.dataset)
    tp_model.train()
    # Set the model to training mode - important for batch normalization and dropout layers
    # Unnecessary in this situation but added for best practices
    for batch, (X, y) in enumerate(dataloader):
        X, y = X.to(device), y.to(device)
        pred = tp_model(X)
        loss = loss_fn(pred, y)
        loss.backward()
        optimizer.step()
        if (rank == 0):
            if batch % 100 == 0:
                loss, current = loss.item(), batch * batch_size + len(X)
                print(f"loss: {loss:>7f}  [{current:>5d}/{size:>5d}]")


from setuptools import setup, find_packages

setup(
    name='detrain',
    version='0.2.5',
    author='A2N Finance', 
    author_email='levi@a2n.finance',
    packages=find_packages(),
    long_description=open('README.md').read()
)


import torch

def test_loop(dataloader, model, loss_fn):
    # Set the model to evaluation mode - important for batch normalization and dropout layers
    # Unnecessary in this situation but added for best practices
    model.eval()
    size = len(dataloader.dataset)
    num_batches = len(dataloader)
    test_loss, correct = 0, 0

    # Evaluating the model with torch.no_grad() ensures that no gradients are computed during test mode
    # also serves to reduce unnecessary gradient computations and memory usage for tensors with requires_grad=True
    with torch.no_grad():
        for X, y in dataloader:
            pred = model(X)
            test_loss += loss_fn(pred, y).item()
            correct += (pred.argmax(1) == y).type(torch.float).sum().item()

    test_loss /= num_batches
    correct /= size
    print(f"Test Error: \n Accuracy: {(100*correct):>0.1f}%, Avg loss: {test_loss:>8f} \n")